/*
**	Command & Conquer Generals(tm)
**	Copyright 2025 Electronic Arts Inc.
**
**	This program is free software: you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation, either version 3 of the License, or
**	(at your option) any later version.
**
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**
**	You should have received a copy of the GNU General Public License
**	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "missingtexture.h"
#include "texture.h"
#include "dx8wrapper.h"
#include <D3dx8core.h>

static unsigned missing_image_width=128;
static unsigned missing_image_height=128;
static unsigned missing_image_depth=24;

extern unsigned int missing_image_palette[];
extern unsigned int missing_image_pixels[];

static IDirect3DTexture8 * _MissingTexture = NULL;

IDirect3DTexture8* MissingTexture::_Get_Missing_Texture()
{
	WWASSERT(_MissingTexture);
	_MissingTexture->AddRef();
	return _MissingTexture;
}

IDirect3DSurface8* MissingTexture::_Create_Missing_Surface()
{
	IDirect3DSurface8 *texture_surface = NULL;
	DX8_ErrorCode(_MissingTexture->GetSurfaceLevel(0, &texture_surface));
	D3DSURFACE_DESC texture_surface_desc;
	::ZeroMemory(&texture_surface_desc, sizeof(D3DSURFACE_DESC));
	DX8_ErrorCode(texture_surface->GetDesc(&texture_surface_desc));
	
	IDirect3DSurface8 *surface = NULL;	
	DX8CALL(CreateImageSurface(
		texture_surface_desc.Width, 
		texture_surface_desc.Height, 
		texture_surface_desc.Format, 
		&surface));
	DX8CALL(CopyRects(texture_surface, NULL, 0, surface, NULL));
	texture_surface->Release();
	return surface;
}

void MissingTexture::_Init()
{
	WWASSERT(!_MissingTexture);

	IDirect3DTexture8* tex=DX8Wrapper::_Create_DX8_Texture(
		missing_image_width,
		missing_image_height,
		WW3D_FORMAT_A8R8G8B8,
		TextureClass::MIP_LEVELS_ALL);

	D3DLOCKED_RECT locked_rect;
	RECT rect;
	rect.left=0;
	rect.right=missing_image_width;
	rect.top=0;
	rect.bottom=missing_image_height;
	DX8_ErrorCode(
		tex->LockRect(
			0,
			&locked_rect,
			&rect,
			0));

	unsigned *buffer=(unsigned*)locked_rect.pBits;
	unsigned char *pixels=(unsigned char *)missing_image_pixels;
	for (unsigned y=0;y<missing_image_height;y++)
	{
		for (unsigned x=0; x<missing_image_width; x++)
		{
			//*buffer++=missing_image_palette[*pixels++];
			*buffer++=0x7FFF00FF;
		}
		buffer=(unsigned*)locked_rect.pBits;
		buffer+=locked_rect.Pitch/sizeof(unsigned)*y;
	}

	DX8_ErrorCode(tex->UnlockRect(0));

	for (unsigned i=1;i<tex->GetLevelCount();++i) {
		IDirect3DSurface8 *src,*dst;
		DX8_ErrorCode(tex->GetSurfaceLevel(i-1,&src));
		DX8_ErrorCode(tex->GetSurfaceLevel(i,&dst));

		DX8_ErrorCode(D3DXLoadSurfaceFromSurface(
			dst,
			NULL,	// palette
			NULL,	// rect
			src,
			NULL,	// palette
			NULL,	// rect
			D3DX_FILTER_BOX,	// box is good for 2:1 filtering
			0));

		src->Release();
		dst->Release();
	}

	_MissingTexture=tex;
/*
	//Load an 8-bit tga and generate text representation
	FILE *fp;
	fp=fopen("missing.tga","rb");
	if (fp)
	{
		char image[128*128];	//make enough storage for image and palette
		char palette[256*3];
		fread(image,18,1,fp);	//skip over the header
		fread(palette,256,3,fp);	//read the palette
		fread(image,1,128*128,fp);
		FILE *output=fopen("missing.txt","w");
		fprintf(output,"palette:\n");
		for (int i=0; i<256; i++)
		{	int color=(((int)palette[i*3+0] & 0x000000ff)|(((int)palette[i*3+1] << 8)&0x0000ff00)|(((int)palette[i*3+2] << 16)&0x00ff0000)) | 0x7f000000;
			fprintf(output,"0x%.8X",color);
			if ((i&7) == 7)	//check for end of 8 element line
				fprintf(output,",\n");	//new line
			else
				fprintf(output,",");	//continue existing line
		}
		fprintf(output,"image:\n");
		for (int y=0; y<128; y++)
		{
			for (int x=0; x<32; x++)
			{	int color=*((int *)(&image[y*128+x*4]));
				fprintf(output,"0x%.8X",color);
				if ((x&7)==7)	//check for end of 8 element line
					fprintf(output,",\n");	//new line
				else
					fprintf(output,",");	//continue existing line
			}
		}
		fclose(output);
		fclose(fp);
	}
*/
}

void MissingTexture::_Deinit()
{
	_MissingTexture->Release();
	_MissingTexture=0;
}

static unsigned int missing_image_palette[]={
0x7F040204,0x7F048AC4,0x7F84829C,0x7FFC0204,0x7F0442AB,0x7FFCFE04,0x7F444244,0x7F0462FC,
0x7F84CEE4,0x7FC4C6CF,0x7F9CA6B2,0x7FC4E6F4,0x7F04FE04,0x7F4C82D4,0x7F2452A1,0x7F0442D4,
0x7F446AB0,0x7FA4A6B6,0x7F2C62C2,0x7FE4E6E9,0x7F646264,0x7F0402FC,0x7FC4D6E1,0x7F44B6DC,
0x7FB4C6D4,0x7F8496A9,0x7F049ECC,0x7F0C52C9,0x7FA4B6C4,0x7FA4DEEC,0x7F0472FC,0x7F3C72C4,
0x7F84A6DF,0x7F748294,0x7F24A6D4,0x7F1442A4,0x7F0452DA,0x7FE4F6FC,0x7FD4D6D6,0x7F849EC9,
0x7F242224,0x7FBCD6F7,0x7FB4B6C2,0x7FD4F6F7,0x7F747274,0x7F0442BF,0x7F4C7ABC,0x7FACC6EC,
0x7F9496AB,0x7F343234,0x7F042A98,0x7F648ADC,0x7F1C5AD4,0x7F545254,0x7F1C62D8,0x7FDCE6E9,
0x7F0452EE,0x7F043AA2,0x7FC4CED3,0x7F2C62D4,0x7F34B6D4,0x7FB4BED9,0x7F6CA6EC,0x7F2C72EC,
0x7F94B6E8,0x7F8482A4,0x7F8CA6CC,0x7F5C72AC,0x7FE4EEEF,0x7FC4D6FC,0x7F1472FC,0x7F7C96C4,
0x7FF4F6F7,0x7FD4DEE3,0x7F6C728C,0x7F949EA4,0x7F5CC6E4,0x7F245ABC,0x7F4472C4,0x7F149ECD,
0x7F6C82BC,0x7F34AED5,0x7F242251,0x7FB4BEC9,0x7F044AC8,0x7FBCCEF0,0x7F1462FC,0x7FD4EEF4,
0x7FC4CEDC,0x7F7C7A97,0x7F0C4AB4,0x7FA4AEBC,0x7F3C62B4,0x7F646284,0x7FBCCEE4,0x7F1C52B7,
0x7F0C5AE2,0x7F043294,0x7F3472E8,0x7F9CB6E1,0x7F74C6E4,0x7F2482FC,0x7F0C42AD,0x7F44426A,
0x7F9CDEEC,0x7F5C9EFC,0x7FB4C6E1,0x7F8C9EAC,0x7FACBED0,0x7F5C7AAC,0x7FB4CEF2,0x7F35325C,
0x7F4C5274,0x7FD4E6F9,0x7F64BEDF,0x7FECEEF2,0x7FDCDEE2,0x7F747A94,0x7F949EB0,0x7F19A6D4,
0x7F342A58,0x7F0C4AC0,0x7F746A8C,0x7F96D6EC,0x7F4C82EC,0x7F345AB0,0x7F044ADA,0x7F4C72B9,
0x7FCCD6E2,0x7F141214,0x7F6496EC,0x7F0696CC,0x7F7C9ECC,0x7F1482FC,0x7F241A4C,0x7FB8E6F4,
0x7F58BEDC,0x7F7CA6E4,0x7F544A74,0x7F5C5A7D,0x7F9CAED0,0x7F443A64,0x7F8CBEFC,0x7F3C6AC9,
0x7F1C5AC8,0x7F848A9C,0x7F8CAEEC,0x7F7C8A9C,0x7F144AB4,0x7F247AFC,0x7FC4DEF4,0x7F6CC6E4,
0x7F2C2A56,0x7F146AF8,0x7F347AF4,0x7F4C4A70,0x7F448AFC,0x7F6C9EFC,0x7F8BD6EC,0x7FC8EEF4,
0x7F2C5AA4,0x7F0A7AFC,0x7FECFEFC,0x7F2C2A2C,0x7F045AF4,0x7F848AA5,0x7F94AED0,0x7FFCFEFC,
0x7F6C8AC4,0x7F646A8A,0x7F3C3A62,0x7F0A6AFC,0x7F9CAEBC,0x7F2C6ACC,0x7FACE6EC,0x7F28AED4,
0x7F043AAF,0x7F2C6ADB,0x7F157AFC,0x7F245AD4,0x7F0432A0,0x7F79CEE4,0x7F7CAEF4,0x7F3C72D8,
0x7F8C96A7,0x7F1452BC,0x7FACB6C3,0x7F7C8299,0x7F0C52D8,0x7F547AC3,0x7FB4C6EC,0x7FCCD6EC,
0x7FBCBEC5,0x7FACAEC0,0x7F94BEF4,0x7F8C8A8E,0x7F1C4AA4,0x7F2C7AEC,0x7F1C6AF0,0x7F3C7AE4,
0x7FCCE6FC,0x7F5482D0,0x7FECE6EC,0x7FBCC6D1,0x7F0C72FB,0x7FECF6F5,0x7FDCF6FA,0x7F0C3AA4,
0x7FCCCECD,0x7F3CB6D7,0x7FBCBED4,0x7FFCF6FC,0x7F74728F,0x7F9C9E9C,0x7F64C6E3,0x7F4C72C5,
0x7F1C9ED0,0x7F3CAED7,0x7FDCEEFC,0x7FCCCEDF,0x7F0C328C,0x7F2C82F4,0x7FBCC6DC,0x7F545275,
0x7FDCE6F8,0x7F9C9EB0,0x7FCCDEEC,0x7F0C5AF0,0x7F8C8AA5,0x7F748ABC,0x7F6C6A8A,0x7F0442B4,
0x7F0C62FC,0x7FC4C6DC,0x7F9CA6BC,0x7F0442E4,0x7F446ABC,0x7F2C62CC,0x7F4CB6DC,0x7F0C9ECC,
0x7FACDEEC,0x7F2CA6D4,0x7F0452E4,0x7FD4D6E4,0x7F849ED4,0x7FB4B6CC,0x7F4C7ACC,0x7FACC6FC,
0x7F9496B4,0x7F042AA4,0x7F1C62E4,0x7F74A6EC,0x7FE4EEFC,0x7F1C72FC,0x7FD4DEEC,0x7F2C5ABC};

static unsigned int missing_image_pixels[]={
0x03030303,0x03030303,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x05050505,0x05050505,
0x03030303,0x03030303,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x05050505,0x05050505,
0x03030303,0x03030303,0xA7A7A7A7,0xD574A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xC3C3C3C0,0xA7A7A7C3,0xA7A7A7A7,0x05050505,0x05050505,
0x03030303,0x03030303,0xA7A7A7A7,0x00C3A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A700A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A700,0xA7A7A7A7,0x00000006,0xA7A70000,0xA7A7A7A7,0x05050505,0x05050505,
0x03030303,0x03030303,0xA7A7A7A7,0x00C3A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A70000,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A70000,0xA7A7A7A7,0xA7A70606,0xA7C300A7,0xA7A7A7A7,0x05050505,0x05050505,
0x03030303,0x03030303,0xA7A7A7A7,0x00C3A7A7,0xA7A7A7A7,0x3174A7A7,0xA33131A3,0x31A33131,
0x3131A331,0xA33131A3,0x31A33131,0x3131A331,0xA33131A3,0x31A33131,0x3131A331,0xA33131A3,
0x31A33131,0x3131A331,0xA33131A3,0x31A33131,0x3131A331,0xA33131A3,0x31A33131,0x3131A331,
0xA33131A3,0x74A33131,0xA7A7A7A7,0xC3C32806,0xA7C00014,0xA7A7A7A7,0x05050505,0x05050505,
0x03030303,0x03030303,0xA7A7A7A7,0x00C3A7A7,0xA7A7A7A7,0x0026A7A7,0x00000000,0x00000000,
0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,
0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,
0x00000000,0x26000000,0xA7A7A7A7,0x00000006,0xA7A72C00,0xA7A7A7A7,0x05050505,0x05050505,
0x03030303,0x03030303,0xA7A7A7A7,0x00C3A7A7,0xA7A7A7A7,0xD0D3A7A7,0xD0D00000,0xD0D0D0D0,
0xD0D0D0D0,0xD0D0D0D0,0xD0D0D0D0,0xD0D0D0D0,0xD0D0D0D0,0xD0D0D0D0,0xD0D0D0D0,0xD0D0D0D0,
0xD0D0D0D0,0xD0D0D0D0,0xD0D0D0D0,0xD0D0D0D0,0xD0D0D0D0,0xD0D0D0D0,0xD0D0D0D0,0xD0D0D0D0,
0xD0D0D0D0,0xD3D00000,0xA7A7A7A7,0x74A70606,0xA7A7D500,0xA7A7A7A7,0x05050505,0x05050505,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x00C3A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A700A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A700,0xA7A7A7A7,0xA7A70606,0xA7A7002C,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x00C3A7A7,0xC3000000,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A70606,0xA70681A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xC0A7A7A7,0xC3C3C3C3,0xA7A7A7C3,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0x14A7A7A7,0x00000000,0xA7A7A700,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA700C3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA700C3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA700C3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA700C3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA700C3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA700C3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x48D026A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x0000A300,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0x00A7A7A7,0x00002800,0xA7A7A700,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x48D348A2,
0xA7A7A7D3,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x7373D3A7,0x44131313,0x73737373,
0x73737373,0xD3CD4873,0xA7A7A7D3,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x1348A7A7,0x2680D026,0x44CA37FE,0xCDCDCD73,
0xCDCDCDCD,0x737348CD,0x26487373,0xA7A748C0,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x2AD226D3,0x26DB3ACB,0x73443774,0x484848CD,
0x48D348D3,0xCDCDCD48,0xFE48CDCD,0xCA2A5902,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x0973A7A7,0x2ABA5B5B,0xFEDB3ACB,0x733774FE,0xCD48A273,
0xA7A2A748,0xCDD3A7A2,0x48CD4848,0x7A30CBBF,0xA7A7CA11,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xF8E10948,0xCBF5C1EA,0x37BF0909,0xF558BFFE,0xA519F8C1,
0x5BF8B819,0xA7A28053,0xCD48A2A7,0xCBF3F374,0xA7C15930,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xF3A7A7A7,0xE1B8E4E1,0xDBD2C15B,0xEADBBF80,0x5B76BBBB,0xC0C0532A,
0x5B2A5309,0x7691A5E1,0xA7A2133D,0xBF588073,0x02E43DF3,0xA7A7A7F3,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0x59C0A7A7,0xC1F8E459,0xBFFE74DB,0x0AD4BB53,0xA748CAD0,0xA7A7A7A7,
0xA7A7A7A7,0xC074D3A7,0x740A4A91,0xF3BFFE48,0xBA80BF58,0xA7A72AE6,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0x59D41173,0xBF5811E4,0xBB7616FE,0xA7730911,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0x4B2ACAA7,0xFEFEBA91,0xF35858BF,0xA7304AD2,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x13A7A7A7,0xB8D45D02,0x58BF583D,0x745B91F8,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x5B937609,0xBFBFFEFE,0x59A95880,0xA7A7A773,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xD426A7A7,0x58EAD48B,0xBBBABF58,0xA7A774D5,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x9109A7A7,0x58FE58B8,0xA980BFBF,0xA7A773E6,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x9B5D74A7,0x74803DD4,0x80BB76F3,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xD3A7A7A7,0x135375BA,0xBFF3DBDB,0xA7485DA9,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x759BA9CA,0x58BF80F5,0xA748B8B8,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xACBBD0A7,0x588080FE,0xA77A7058,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0x48A7A7A7,0x5859AAD4,0x2158BF80,0xA7A7A711,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x9173A7A7,0x80BFFEB8,0xE4AAD280,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xE4A7A7A7,0xF3807598,0xC07558FE,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x48A7A7A7,0x80FEA530,0x9811FEBF,0xA7A7A7C1,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0x5211A7A7,0x49F3585D,0xA709BB58,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xFE6BE1D3,0x7549BFBF,0xA7A7DB98,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0x676F74A7,0xFEBF803D,0xA7A7BABB,0xA7A7A7A7,0x48A7A7A7,0x48484848,0x48484848,0x48484848,
0x48484848,0xA7A7A748,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xF891A7A7,0xE9FE8074,0xA7A78B6F,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xEA985DA7,0xA58058F3,0xA7A7A70A,0xA7A7A7A7,0x8F55A7A7,0x4ED74ED7,0x4ED74ED7,0x4ED74ED7,
0x8F7F4ED7,0xA7A748A8,0xA7A77463,0xA7A7A7A7,0xA7A7A7A7,0xBB73A7A7,0x8058FEBA,0xA711520A,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xFEE4862A,0xBBEABF58,0xA7A7A7D3,0xA7A7A7A7,0xB45CCDA7,0xB0B039F9,0x39B0B0B0,0xB0B039B0,
0xF9B0B039,0x6EA7A742,0xA7A75039,0xA7A7A7A7,0xA7A7A7A7,0x49A7A7A7,0xDB745875,0x73AA8BFE,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x73A7A7A7,
0xBF80DF9B,0x13BB5880,0xA7A7A7A7,0xA7A7A7A7,0x043984A7,0xE7040404,0xE7545AE7,0x5A2D5A54,
0xBDB05A54,0x1FCDA7A7,0xA7DE04F9,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xBFFEB8C0,0xE452BA80,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x11A7A7A7,
0x80F3BA86,0xA7539337,0xA7A7A7A7,0xA7A7A7A7,0x39B404DB,0xB0B0B0B0,0xB0B0B0B0,0xB0B0B0B0,
0x7494B0B0,0xB089A7A7,0x737D3904,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xFEACB8A7,0x985DBFF3,
0xA7A7A7CA,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xDF48A7A7,
0xBF58808B,0xA7D3913D,0xA7A7A7A7,0xA7A7A7A7,0xBDBD7F8C,0xBDC9BDBD,0xBDBDC9BD,0xC9BDBDC9,
0xA7A6D7BD,0x041B96A7,0x47390404,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x809126A7,0x98E1F3BF,
0xA7A7A711,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x98F5A7A7,
0xBFBFBFBA,0xA7A726BB,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x5AB00DA2,0x6639045A,0xA7A7A780,0xA7A7A7A7,0xA7A7A7A7,0xEA0AA7A7,0x7058BFF3,
0xA7A773DF,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x8B5DA7A7,
0xBA74BFFE,0xA7A7D3A5,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x5A542D40,0xB0E75A5A,0xA7A7485C,0xA7A7A7A7,0xA7A7A7A7,0x91F3A7A7,0x59BFF3BF,
0xA7A72A6F,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x116F26A7,
0xB8BF80FE,0xA7A7A7D0,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xDAA7A7A7,0x545A2D90,0x045A5A2D,0xA7A7A639,0xA7A7A7A7,0xA7A7A7A7,0xE1A7A7A7,0xBAFEFEBA,
0xA7A7E667,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x589BE1A7,
0xBBDEDB58,0xA7A7A773,0xA7A7A7A7,0xFCDACDA7,0xFCFCFCFC,0xDAFCFC25,0xFCFCFCFC,0xA7CDE073,
0x82A7A7A7,0x541B247E,0x5A2D544D,0xA7E00EB0,0xA7A7A7A7,0xA7A7A7A7,0x09A7A7A7,0xD2BFFEBB,
0xA7FEDF8B,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xFE41E6A7,
0x76F8DBDB,0xA7A7A7A7,0xA7A7A7A7,0x124DD773,0xED12124D,0xB1B1ADB3,0xB1B1B1B1,0xA7553BC6,
0x246EA7A7,0xED54241B,0x5AB0BDE2,0xA7A8F9E7,0xA7A7A7A7,0xA7A7A7A7,0xD3A7A7A7,0xBFFE5875,
0xA7C09BD4,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x80CB8B73,
0xD0BB5858,0xA7A7A7A7,0xA7A7A7A7,0x39B43227,0xE7B039B0,0x542DE72D,0x3824547E,0xA74833EB,
0x383FDAA7,0x637EBCA4,0xE7B9BFA7,0x6C39E75A,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xBFFEF811,
0xA7E170B8,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xF3FE5D2A,
0xD3BB58DB,0xA7A7A7A7,0x5EA7A7A7,0x66043966,0x5A545AE7,0x24245454,0x38A4A460,0xA7A7733F,
0xA4389DA7,0xD3C97EA4,0x2D20A7A7,0x0EB05A5A,0xA7A7A737,0xA7A7A7A7,0xA7A7A7A7,0xFEFEBB26,
0x73598B0A,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x80FEBBE4,
0xA74B6CFE,0xA7A7A7A7,0xA0FCA7A7,0x04043932,0x54545AE7,0x7E540F2D,0x3838387E,0xA7A7A763,
0x56E8E845,0xA77156A4,0xD7A7A7A7,0x39045A2D,0xA7A7A7E5,0xA7A7A7A7,0xA7A7A7A7,0x805891D3,
0x13E65DF5,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xDE58F859,
0xA7C0EABF,0xA7A7A7A7,0x4713A7A7,0x045C42F4,0x5F2D5A5A,0x202020F4,0x40202020,0x48A7A7D3,
0x07AB079C,0xA7A792A4,0xF6FB89CD,0x04E75A2D,0xA7A7E966,0xA7A7A7A7,0xA7A7A7A7,0xFE3D0AA7,
0xF3E6D4DE,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xDBDBBA59,
0xA7F3B8BF,0xA7A7A7A7,0xA7A7A7A7,0x04FF48A7,0xF40454E7,0xA7A7A7A7,0xA7A7A7A7,0xC2A7A7A7,
0x0746CC1E,0xA7A7487C,0x542D0FB6,0x045A5A5A,0xA7485C39,0xA7A7A7A7,0xA7A7A7A7,0xFE5BBAA7,
0xD0E6BB58,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x73A7A7A7,0xDBE9DEE4,
0xA713BBBF,0xA7A7A7A7,0xA7A7A7A7,0xE7B0A8A7,0xD34E2D5A,0xA7A7A7A7,0xA7A7A7A7,0xB2C8A7A7,
0xABCCCCCC,0x29A7A745,0x5424BCA4,0x66E75454,0xA7276104,0xA7A7A7A7,0xA7A7A7A7,0xFEF8C0A7,
0xCBD4E458,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xCAA7A7A7,0xBF585830,
0xA7D3BB58,0xA7A7A7A7,0xA7A7A7A7,0xE704CF5E,0xA7805A66,0xA7A7A7A7,0xA7A7A7A7,0x1E69A7A7,
0x3EABA1A1,0x9A48A7A7,0x0F7EF27E,0x0404E72D,0x74943904,0xA7A7A7A7,0xA7A7A7A7,0xF3B8D0A7,
0xF5BBE480,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x74A7A7A7,0xF358DEB8,
0xA7D3BB58,0xA7A7A7A7,0x48A7A7A7,0xE766398F,0xC8C80DB0,0xC8C8C80B,0xC871C8C8,0xA1858E71,
0x73DDA185,0x9AF7A7A7,0x8FC7B79A,0x04B98F8F,0x43610466,0xA7A7A7A7,0xA7A7A7A7,0xF3A580A7,
0x2ABBF858,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xF3A7A7A7,0xDBE9D2B8,
0xA7A7BBDE,0xA7A7A7A7,0x84A7A7A7,0xE7660432,0x3490905A,0x99993636,0x65B2FD46,0x85A1A165,
0xA72F1E85,0xFC48A7A7,0x73737373,0x5F554873,0x61663904,0xA7A7A7AC,0xA7A7A7A7,0xBFA526A7,
0x5302E1BF,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x26A7A7A7,0x58DBE9F8,
0xA7A79118,0xA7A7A7A7,0x0480A7A7,0xE7040439,0x542D2D5A,0xA4F2247E,0x1EAB07E8,0x8585A1CC,
0xA7A73E1E,0xA7A7A7A7,0xA7A7A7A7,0xA6A7A7A7,0x390404B0,0xA7A7370E,0xA7A7A7A7,0xF39149A7,
0xC0E4E109,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xF3A7A7A7,0xBF585830,
0xA7A741D2,0xA7A7A7A7,0x321048A7,0xE7040439,0x54542DE7,0xA4A42424,0x1EABAB07,0xA1A1A1A1,
0xA7A71365,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xB439B42E,0xA7A76D61,0xA7A7A7A7,0x80A5F3A7,
0xD2301158,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x74A7A7A7,0xBF585830,
0xA7D3BB58,0xA7A7A7A7,0x3932A6A7,0x04393939,0x542DE704,0xE3242454,0xABABE8E3,0xCC1ECCCC,
0xA7A7A7C2,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x39B46616,0xA76ADC61,0xA7A7A7A7,0xFEB8D0A7,
0xD0E111D2,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xCAA7A7A7,0xBF58D2B8,
0xA773BB13,0xA7A7A7A7,0x6A3DF3A7,0xBE6A6A6A,0xBE6ABE6A,0x6E6EBE6A,0x6E6E6E6E,0x456E6E55,
0xA7A7A748,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x6A3DBFA7,0xA7746C6A,0xA7A7A7A7,0xFEA509A7,
0xDB11C13D,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x48A7A7A7,0x7358D2B8,
0xA713B8A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA2A7A7A2,0xA248A2A2,0xA7A7A2CD,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xFE0AC0A7,
0x74C12AE9,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA237BA91,
0xA7D00AA7,0xA7A7A7A7,0x4F77B5A7,0x77EF77EF,0xA7A7A2D6,0x0B77D6A7,0x48A7A7A7,0xA7A764F1,
0xF0A7A7A7,0x770B0B77,0x77777777,0xA2CD51EF,0xAFAF7797,0x9F873C77,0xA7F07B9E,0xFE1CC1A7,
0xCAF553C0,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xCDD358E1,
0xA7BA5BA7,0xA7A7A7A7,0x88774F71,0x77888888,0xA7A78701,0xEE831A87,0x25A7A7A7,0xA7AE0177,
0xEFA2A7A7,0x83878783,0x6464B551,0x7BA20897,0x6464D101,0x2BD80197,0xA7F0689F,0x58DE30A7,
0x73C0D2CB,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xCDCD7373,
0xA7B858A7,0xA7A7A7A7,0xA7648387,0x57A7A7A7,0xA2A75797,0x1AB52288,0x25A7A70B,0xA25183AF,
0x8397A7A7,0x8387871A,0xA7A7A77B,0xB5A7A7A7,0xA7A78783,0xCD887BA7,0xA7A7A7A7,0x26809148,
0xD3580958,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xCD48CDA7,
0x73934448,0xA7A7A7A7,0xA7971A0B,0x878787A7,0x9FA7A287,0x1AA77B83,0x25A7A797,0x647797AF,
0x9E1A1DA7,0x8387871A,0x7B7297F1,0x9EA7A7A7,0x9E7B8883,0xA7251D9E,0xA7A7A7A7,0xFEFEBB09,
0xA7F3D080,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x48CD48A7,
0xC0F8A748,0xA7A7A7A7,0xA797EF87,0x1A4F22CE,0xD6A78783,0x88A2DAD8,0xCDA70B83,0xEF7B8777,
0x9F972287,0x1A87871A,0x88EF771A,0xA2A7A7A7,0xEF4F4F88,0x25D9834F,0xA7A7A7A7,0xF337EAB8,
0xA7138026,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xCD7348A7,
0xB8BAA248,0xA7A7A7A7,0xA7641A0B,0x649FDAA7,0x1A258701,0x7717D91A,0x25A7D601,0xAFA7F077,
0x0BCDAF83,0x8387871A,0xA7A2A708,0xDAA7A7A7,0x25CECE9F,0x571A1757,0x73A7A7A7,0x74F3BF75,
0xA7D349FE,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xCD73A7A7,
0xBB58A7CD,0xA7A7A713,0xA2D6EF87,0x08A7A2CD,0x83B59F01,0xD951D917,0x25CE1A4F,0x87A71D77,
0x87A76801,0x8387AE1A,0xA2A2A264,0xAFA7A2CD,0xA7A7A7F1,0x2B8388A7,0x53A7A7A7,0x1374FE76,
0xA7A73774,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x7373A7A7,
0xB848A248,0xA7A7A753,0x228377CE,0x4FF12222,0x88D81DEF,0xA2A2A2A2,0x2564017B,0xA7A77BEF,
0x87A74868,0x83878783,0x2222F14F,0x88CED977,0x3C3C5101,0xCD7783F1,0x30A7A7A7,0x1373FE53,
0xA7A77313,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x44D3A7A7,
0xDEA7CD48,0xA7A77375,0x88721DA7,0x88888888,0x9F72871D,0xA7A7A7A7,0xCD64B5A7,0xA7A78797,
0xDAA7A7A7,0xD6DA574C,0x88888888,0xCD25D688,0x888888B5,0xA70BD688,0xBB09A7A7,0x447373BF,
0xA7A7A713,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xCDA7A7A7,
0x48A24873,0xA7A7C0BB,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x5BB8A7A7,0x4448CD48,
0xA7A7A748,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xD3A7A7A7,
0xA248CD73,0xA748D4DB,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xE07509A7,0x73CD48A2,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA248CD73,0xA75BB848,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7BABB48,0x48734848,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xCD4873D3,0xD075F3A2,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA248E40A,0xA273CD48,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xCD7373A7,0x59F5FC48,0xA7A7A737,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xC0A7A7A7,0xCDA2FE4A,0xA7487348,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0x4873D3A7,0xBA74CD48,0xA7A748B8,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x7526A7A7,0xCD48A2DE,0xA7A773CD,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0x7348A7A7,0xFEE07373,0xA7A730B8,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xBA7573A7,0xCD4848A7,0xA7A7D373,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0x44A7A7A7,0xF3134444,0xA7E1A5BF,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7CBBB73,0x73CD48A2,0xA7A7A748,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xD3A7A7A7,0xFE3713E0,0x307680FE,0xA7A7A7D3,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x74A7A7A7,0x48A73DBB,0x4873CDCD,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x49747473,0xA580FE49,0xA7A713E4,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xBB3AA7A7,0xCDD3A7BA,0xA7737348,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x26F313A7,0x74DB80DB,0xA7DB755B,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xFE2111A7,0xCDCDD3A7,0xA7A773CD,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xDB74A7A7,0x80CB093A,0xBABBBAFE,0xA7A7A748,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x13A7A7A7,0xA737194B,0x48CDCDD3,0xA7A74873,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x49A7A7A7,0xD2C0D2CB,0xB8BFFEF3,0xA7A73A59,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x4AC173A7,0x13E0FE5B,0x4473CD73,0xA7A7A7D3,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0x2AF55374,0xFEBFD211,0x0902A5DB,0xA7A7A773,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0x74A7A7A7,0xBFB875C1,0x5880DBFE,0x74E4D2BF,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0x5BBA74A7,0xF3F5115B,0xB86C80FE,0xA726E1BB,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0x93B8CB48,0xFEFEE90A,0xFE58F3BF,0xA7C04A11,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xBACAA7A7,0x5B30E1E1,0xFEFEBFE9,0x59BBEAD2,0x48135811,0xA7A7A7A7,
0xA7A7A7A7,0xE1CB7473,0xBF5B9175,0x58F3FEFE,0xE4585858,0xA7A72659,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xD3A7A7A7,0xE4B83009,0xBF58C1E4,0x5858FEFE,0x21930ABA,0xB8E491BB,
0xBB91E4E4,0xF5EAA5BB,0xBFFEFE58,0x80585858,0x1175BADB,0xA7A7A7D3,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xE45B13A7,0xEAA5BBBB,0xFEFEF309,0x5880FEFE,0xBABAF5DE,
0xDB3DC11C,0xBF74FEBF,0xDBDBDBBF,0xC180FEF3,0xA726B802,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x26A7A7A7,0xE6E65911,0x585B3059,0xF3FE49BF,0xF3FEF3FE,
0xF3FEBFFE,0x5858BF80,0xFEBF8058,0x3041B8CB,0xA7A7A7D2,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xD411F3A7,0xD48B8B8B,0x80DEC1B8,0xF3FEF3BF,
0xDBBFF3F3,0x49FE80BF,0xE630CBF3,0xD3D0915D,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xCAA7A7A7,0x9BE6E42A,0xE6DF6767,0x0AF8A559,
0xEAEA5BEA,0xA9BBE4F8,0x11595D8B,0xA7A7A7CA,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xC073A7A7,0x8A5D5930,0x986F6F8D,
0xAA6F6F78,0x025DDF67,0xA7A7732A,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x73D3A7A7,0xD209D074,
0x74D009C0,0xA7A7A773,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x2600A3A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0x00A7A7A7,0x00002800,0xA7A7A700,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x0000A300,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0x743135A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0x74A7A7A7,0xC0C3C3C3,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xC3A7A7A7,0x00000000,0xA7A7A731,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xC3A7A7A7,0x74A7A700,0xA7A7A700,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xC3A7A7A7,0x35C3C300,0xA7A7A700,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xC3A7A7A7,0x00000000,0xA7A7A7A3,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xC3A7A7A7,0xA7A7A700,0xA7A7D500,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xC3A7A7A7,0xC0A7A700,0xA7A7D500,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xC3A7A7A7,0x00000000,0xA7A7A706,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0x0C0C0C0C,0x0C0C0C0C,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x15151515,0x15151515,
0x0C0C0C0C,0x0C0C0C0C,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x15151515,0x15151515,
0x0C0C0C0C,0x0C0C0C0C,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x15151515,0x15151515,
0x0C0C0C0C,0x0C0C0C0C,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x15151515,0x15151515,
0x0C0C0C0C,0x0C0C0C0C,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x15151515,0x15151515,
0x0C0C0C0C,0x0C0C0C0C,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x15151515,0x15151515,
0x0C0C0C0C,0x0C0C0C0C,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x15151515,0x15151515,
0x0C0C0C0C,0x0C0C0C0C,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,
0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0xA7A7A7A7,0x15151515,0x15151515};
